/*	Simple Web Server
 Written by Dimitar Kovachev, f.#. 42681        */

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

class WebServerThread extends Thread {
    private Socket socket;
    private BufferedReader in;
    private PrintWriter out;
	private String strRootDir;
	
    public WebServerThread(Socket s, String webServerRootDirectory)
		throws IOException
	{
		strRootDir = webServerRootDirectory;
        socket = s;
        in =
            new BufferedReader(
                new InputStreamReader(
                    socket.getInputStream()));
        // Enable auto-flush:
        out =
            new PrintWriter(
                new BufferedWriter(
                    new OutputStreamWriter(
                        socket.getOutputStream())), true);
        // If any of the above calls throw an
        // exception, the caller is responsible for
        // closing the socket. Otherwise the thread
        // will close it.
        start(); // Calls run()
    }
    public void run() {
		try {
			
			// Get the request, send the reply
			String strRequest = getClientRequest();
			writeServerResponse(strRequest);
			
        } catch (IOException e) {
        } finally {
            try {
                socket.close();
            } catch(IOException e) {}
        }
    }
	private String getClientRequest()
		throws IOException
	{
		// Get the first line of the request - it's the only one we need
		String strFirstLine = in.readLine();
		
		// Read all the lines, until a blank one is reached. Ignore them.
		String strNextLine = strFirstLine;
		while (!strNextLine.equals(""))
		{
			strNextLine = in.readLine();
		}
		
		// Return the first line of the request
		return strFirstLine;
	}
	
	private void writeServerResponse(String strRequest)
		throws IOException
	{
		StringTokenizer tokenizer = new StringTokenizer(strRequest);
		
		String strCommand;
		String strURL;
		String strProtocol;
		
		// Check if the request has enough components
		if (tokenizer.countTokens() < 3)
		{
			writeNotSupportedRequestError();
			return;
		}
		else
		{
			strCommand 	= tokenizer.nextToken();
			strURL 		= tokenizer.nextToken();
			strProtocol = tokenizer.nextToken();
		}
		
		// Check if the command is GET
		if (!strCommand.equals("GET"))
		{
			writeNotSupportedRequestError();
			return;
		}
		
		// Check the protocol?
		
		// Handle URL-formed requests. Convert to filename.
		try
		{
			URL targetURL = new URL(strURL);
			strURL = targetURL.getFile();
		}
		catch (MalformedURLException e)
		{
			// The reques is not an URL, we don't have to change it
		};
		
		// Get the file, requested by the client
		File targetFile = new File (strRootDir + strURL);
		System.out.println("Requested : " + targetFile);
		
		// If it don't exists - return error 404
		if (!targetFile.exists())
		{
			System.out.println("File " + targetFile + " not found!\n");
			writeFileNotFoundError();
			return;
		}
		
		// If we don't have the rights to read it - return error 403
		if (!targetFile.canRead())
		{
			System.out.println("File " + targetFile + " can't be read!\n");
			writeAccessDeniedError();
			return;
		}
		
		// If it is a directory : List the contents
		if (targetFile.isDirectory())
		{
			writeDirectoryContents(strURL);
			return;
		}
		
		// The client has requested a file. Send it.
		writeSimpleFile(targetFile);
	}
	
	private void writeNotSupportedRequestError()
	{
		out.print(	"HTTP/1.0 501 Not Implemented\n" +
					"MIME-Version: 1.0\n" +
					"Content-type: text/html\n" +
					"\n" +
					"<html> Not implemented </html>\n");
	}
	
	private void writeFileNotFoundError()
	{
		out.print(	"HTTP/1.0 404 Not Found\n" +
					"MIME-Version: 1.0\n" +
					"Content-type: text/html\n" +
					"\n" +
					"<html> Requested document not found! </html>\n");
	}
	
	private void writeAccessDeniedError()
	{
		out.print(	"HTTP/1.0 403 Forbidden\n" +
					"MIME-Version: 1.0\n" +
					"Content-type: text/html\n" +
					"\n" +
					"<html> Requested document can not be accessed! </html>\n");
	}
	
	private void writeDirectoryContents(String strURL)
	{
		// Get the canonical form of the directory's name
		File a = new File (strURL);
		strURL = a.getPath();
		
		// Override system-dependant slash type
		strURL = strURL.replace('\\', '/');
		
		if (strURL.equals("//"))
			strURL = "/";
		
		File targetDir = new File (strRootDir + strURL);
		String[] allFilesInDir = targetDir.list();

		// Add a final slash to the directory name, if it is missing
		if (!strURL.endsWith("/"))
		{
			strURL += '/';
		}
		
		// Print the header of the response
		writeSuccessHeader("text/html\n");
		
		// Print the opening <HTML> tag
		out.println("<HTML>\n");
		
		// Report the current directory
		out.println("Directory listing of " + getFileWithLink(strURL) + " <br>\n");
		
		// Add link to the parent directory (unless we are in the root)
		if (!strURL.equals("/"))
			out.println("\t" + getFileWithLink1(strURL, "..") + "\n");
		
		// Print all files in the directory
		for (int i = 0; i < allFilesInDir.length; i++)
			out.println("\t" + getFileWithLink1(strURL, allFilesInDir[i]) + "\n");
		
		// Print the closing </HTML> tag
		out.println("</HTML>\n");
	}
	
	private void writeSimpleFile(File targetFile)
		throws IOException
	{
		String[] fileExtensions = {
				".htm",
				".html",
				".gif",
				".jpg",
				".jpeg",
				".zip",
				
				""				// Handle all unknown file-types
			};
		
		String[] contentTypes = {
				"text/html",
				"text/html",
				"image/gif",
				"image/jpg",
				"image/jpg",
				"application/zip",
				
				"text/plain"	// Handle all unknown file-types
			};
				
		// Get the filename in lower-case
		String lowerCaseFileName = targetFile.getName().toLowerCase();
		
		// Determine the content-type and write the header
		for (int i = 0; i < fileExtensions.length && i < contentTypes.length; i++)
			if (lowerCaseFileName.endsWith(fileExtensions[i]))
			{
				writeSuccessHeader(contentTypes[i]);
				break;
			}
		
		// Send the file
		FileInputStream fileReader;
		try
		{
			// Create an input stream
			fileReader = new FileInputStream(targetFile);
		} catch (FileNotFoundException e)
		{		// Do nothing - we know that the file exists
			return;
		}
		
		// Create an output stream
		DataOutputStream outWriter = new DataOutputStream(socket.getOutputStream());
		
		// Transfer data until the input stream is empty
		byte[] dataBuffer = new byte[1024];
		while (true)
		{
			int bytesRead = fileReader.read(dataBuffer);
			if (-1 == bytesRead)
				break;
			outWriter.write(dataBuffer, 0, bytesRead);
		}
		
		// Close the input stream, flush the output stream
		fileReader.close();
		outWriter.flush();
	}
	
	private String getFileWithLink(String fileName)
	{
		return "<a href=\"" + fileName + "\">" + fileName + "</a> <br>";
	}
	
	private void writeSuccessHeader(String contentType)
	{
		out.print(	"HTTP/1.0 200 OK\n" +
					"MIME-Version: 1.0\n" +
					"Content-type: " + contentType + "\n" +
					"\n");
	}
	
	private String getFileWithLink1(String baseURL, String fileName)
	{
		return "<a href=\"" + baseURL + fileName + "\">" + fileName + "</a> <br>";
	}
}

public class SimpleWebServer
{
    static final int PORT = 80;
    public static void main(String[] args)
		throws IOException
	{
        ServerSocket s = new ServerSocket(PORT);
		System.out.println("Simple Web Server started on port 80...");
		
		String webServerRootDirectory = "C:/";
		if (args.length > 0)
		{
			webServerRootDirectory = args[0];
		}
		
		System.out.println("Web Server's main directory is " + webServerRootDirectory + "\n");
			
        try {
            while(true) {
                // Blocks until a connection occurs:
                Socket socket = s.accept();
                try {
                    new WebServerThread(socket, webServerRootDirectory);
                } catch(IOException e) {
                    // If it fails, close the socket,
                    // otherwise the thread will close it:
                    socket.close();
                }
            }
        } finally {
            s.close();
        }
    }
}


