import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * Mailbox class implementing mail server mailbox for a partucular user
 */
class Mailbox {
	String username;
	Vector messages;

	/**
	 * Constructor
	 *
	 * @param username - The username to whom this mailbox belongs
	 */
	public Mailbox(String username) {
		this.username = username;
		messages      = new Vector();
	}

	/**
	 * Stores a message in the user's mailbox
	 *
	 * @param msg - The text of the messsage
	 */
	public synchronized void putMessage(String msg) {
		messages.add(msg);
	}

	/**
	 * Retrieves the first message from the user's mailbox
	 *
	 * @return The text of the messsage
	 */
	public synchronized String getMessage() {
		return (String)messages.firstElement();
	}

	/**
	 * Retrieves the i-th message from the user's mailbox
	 *
	 * @param i - The message index
	 *
	 * @return The text of the messsage
	 */
	public synchronized String getMessage(int i) {
		return (String)messages.elementAt(i);
	}

	/**
	 * Gets the number of the messages inside the mailbox
	 *
	 * @return The number of the messsages
	 */
	public synchronized int getCount() {
		return messages.size();
	}
}

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * SMTPThread used for comunication with the client via SMTP
 */
class SMTPThread extends Thread {
	static final String HELO 		= "HELO";
	static final String MAIL_FROM	= "MAIL FROM:";
	static final String RCPT_TO		= "RCPT TO:";
	static final String DATA		= "DATA";
	static final String QUIT		= "QUIT";
	static final String UNKNOWN		= "UNKNOWN";

	HashMap users;
	BufferedReader in;
	PrintWriter out;
	Socket socket;

	/**
	 * Constructor - opens input/output streams and starts the thread
	 *
	 * @param s - Socket used for comunication
	 * @param users - Users' mailboxes
	 *
	 * @throws IOException
	 */
	public SMTPThread(Socket s, HashMap users)
		throws IOException {
		this.users = users;
		socket = s;

		// get input stream
        in = new BufferedReader(
                new InputStreamReader(
                    socket.getInputStream()));

		// get output stream
	    out = new PrintWriter(
                new BufferedWriter(
                    new OutputStreamWriter(
	                   socket.getOutputStream())), true);

		// start the thread
		start();
	}

	/**
	 * The SMTP thread lifecycle
	 */
	public void run() {
		String line;		// the current line
		String cmd;			// the current command that is send from the client
		String expectedCmd; // the command that the client is expected to send
		String email = "";

		try {
			// start talking
			out.println("220 Simple Mail Server");
			expectedCmd = MAIL_FROM;

			// read the input
			while (true) {
				// get one line
				if ((line = in.readLine()) == null)
					break;

				// get the command from this line
				cmd  = getCommand(line);

				// now check the command
				if (HELO.equals(cmd)) {
					// salute the client
					out.println("250 Welcome " + line.substring(HELO.length()).trim());
					expectedCmd = MAIL_FROM;
				} else if (QUIT.equals(cmd)) {
					// go out
					out.println("221 Bye");
					break;
				} else if (UNKNOWN.equals(cmd)) {
					// what's this
					if (!"".equals(line.trim()))
						out.println("500 Invalid command");
				} else if (MAIL_FROM.equals(cmd) && MAIL_FROM.equals(expectedCmd)) {
					// sending mail from
					out.println("250 " + getEmail(line, cmd) + " Sender OK");
					expectedCmd = RCPT_TO;
				} else if (RCPT_TO.equals(cmd) && RCPT_TO.equals(expectedCmd)) {
					// sending mail to
					email = getEmail(line, cmd);

					// check if this user is valid
					if (users.containsKey(email)) {
						out.println("250 " + email + " Recipient OK");
						expectedCmd = DATA;
					} else {
						out.println("550 " + email + " Recipient not found on this server");
					}
				} else if (DATA.equals(cmd) && DATA.equals(expectedCmd)) {
					// now sending the data
					out.println("354 Enter mail, end with \".\" on a line by itself");
					StringBuffer msg = new StringBuffer();
					
					// read the message line by line and append it to the buffer
					while (true) {
						line = in.readLine();

						if (".".equals(line)) {
							// this is the end of the message, store it in the mailbox
							((Mailbox)users.get(email)).putMessage(msg.toString());
							break;
						} else
							msg.append(line + "\n");
					}
					out.println("250 Message delivered");
					expectedCmd = MAIL_FROM;
				} else {
					out.println("503 Invalid state");
				}
			}
        } catch (IOException e) {
        } finally {
            try {
                socket.close();
            } catch(IOException e) {}
        }
	}

	/**
	 * Parses the line and get the command
	 *
	 * @param line - The line that was sent by the client
	 *
	 * @return The command
	 */
	private String getCommand(String line) {
		if (line != null) {
			if (line.toUpperCase().startsWith(HELO))
				return HELO;
			if (line.toUpperCase().startsWith(MAIL_FROM))
				return MAIL_FROM;
			if (line.toUpperCase().startsWith(RCPT_TO))
				return RCPT_TO;
			if (line.toUpperCase().startsWith(DATA))
				return DATA;
			if (line.toUpperCase().startsWith(QUIT))
				return QUIT;
		}
		return UNKNOWN;
	}

	/**
	 * Retrives the email from MAIL FROM and RCPT TO commands
	 *
	 * @param line - The line that was sent by the client
	 * @param cmd  - The command from this line
	 *
	 * @return The email
	 */
	private String getEmail(String line, String cmd) {
		line = line.substring(cmd.length()).trim();
		if (line.startsWith("<") && line.endsWith(">"))
			line = line.substring(1, line.length()-1);
		return line;
	}
}

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * SMTPServer listens on port 25 for connections and starts new SMTPThread for each incomming request
 */
class SMTPServer extends Thread {
	static final int PORT = 25;
	HashMap users;
	ServerSocket s;

	/**
	 * Constructor - opens the socket
	 *
	 * @param users - Users' mailboxes
	 */
	public SMTPServer(HashMap users) {
		this.users = users;
		try {
			// open the port
			s = new ServerSocket(PORT);
			start();
		} catch (IOException e) {}
	}

	/**
	 * The SMTP server lifecycle
	 */
	public void run() {
	  	try {
			// listen for incomming connections
  			while (true) {
				Socket socket = s.accept();
				try {
					new SMTPThread(socket, users);
				} catch (IOException e) {
					socket.close();
				}
			}
		} catch (IOException e) {
		} finally {
			try {
                s.close();
            } catch(IOException e) {}
		}
	}
}

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * POP3Thread used for comunication with the client via POP3
 */
class POP3Thread extends Thread {
	static final String USER 		= "USER";
	static final String PASS		= "PASS";
	static final String LIST		= "LIST";
	static final String RETR		= "RETR";
	static final String QUIT		= "QUIT";
	static final String UNKNOWN		= "UNKNOWN";

	HashMap users;
	BufferedReader in;
	PrintWriter out;
	Socket socket;

	/**
	 * Constructor - opens input/output streams and starts the thread
	 *
	 * @param s - Socket used for comunication
	 * @param users - Users' mailboxes
	 *
	 * @throws IOException
	 */
	public POP3Thread(Socket s, HashMap users)
		throws IOException {
		this.users = users;
		socket = s;

		// get input stream
        in = new BufferedReader(
                new InputStreamReader(
                    socket.getInputStream()));

		// get output stream
		out = new PrintWriter(
                new BufferedWriter(
                    new OutputStreamWriter(
	                   socket.getOutputStream())), true);

		// launch the thread
		start();
	}

	/**
	 * The POP3 thread lifecycle
	 */
	public void run() {
		String line;		// the current line
		String cmd;			// the current command send by the client
		boolean logged;		// store here the client state
		String email = "";

		try {
			// start talking
			out.println("+OK POP3 server ready");
			logged = false;

			// read the input
			while (true) {
				// read one line
				if ((line = in.readLine()) == null)
					break;

				// get the command
				cmd  = getCommand(line);

				// now check the command
				if (USER.equals(cmd) && !logged) {
					// client send USER
					email = getEmail(line, cmd);
					out.println("+User OK, please send your password");
				} else if (QUIT.equals(cmd)) {
					// astalavista
					out.println("+ Bye");
					break;
				} else if (UNKNOWN.equals(cmd)) {
					// ??? WHAT
					if (!"".equals(line.trim()))
						out.println("-ERR Invalid command");
				} else if (PASS.equals(cmd) && !"".equals(email) && !logged) {
					// sending PASS
					// check if the user exist and the password matches
					if (email.equals(getEmail(line, cmd)) && users.containsKey(email)) {
						out.println("+OK");
						logged = true;
					} else {
						email = "";
						out.println("-ERR Invalid username/password");
					}
				} else if (LIST.equals(cmd) && logged) {
					// give the number of the messages
					int count = ((Mailbox)users.get(email)).getCount();
					out.println("+OK " + count + " messages");
					for (int i = 0; i < count; i++) {
						out.println((i + 1) + " " + ((Mailbox)users.get(email)).getMessage(i).length());
					}
					out.println(".");
				} else if (RETR.equals(cmd) && logged) {
					// send the requested message
					try {
						int i = Integer.parseInt(getEmail(line, cmd))-1;
						if (i > -1 && i < ((Mailbox)users.get(email)).getCount()) {
							out.println("+OK message " + (i+1));
							out.println(((Mailbox)users.get(email)).getMessage(i));
						} else
							out.println("-ERR No such message");
					} catch (Exception e) {
						out.println("-ERR No such message");
					}
				} else {
					out.println("-ERR Invalid state");
				}
			}
        } catch (IOException e) {
        } finally {
            try {
                socket.close();
            } catch(IOException e) {}
        }
	}

	/**
	 * Parses the line and get the command
	 *
	 * @param line - The line that was sent by the client
	 *
	 * @return The command
	 */
	private String getCommand(String line) {
		if (line != null) {
			if (line.toUpperCase().startsWith(USER))
				return USER;
			if (line.toUpperCase().startsWith(PASS))
				return PASS;
			if (line.toUpperCase().startsWith(RETR))
				return RETR;
			if (line.toUpperCase().startsWith(LIST))
				return LIST;
			if (line.toUpperCase().startsWith(QUIT))
				return QUIT;
		}
		return UNKNOWN;
	}

	/**
	 * Retrives the user from USER and PASS commands
	 *
	 * @param line - The line that was sent by the client
	 * @param cmd  - The command from this line
	 *
	 * @return The email
	 */
	private String getEmail(String line, String cmd) {
		return line.substring(cmd.length()).trim();
	}
}

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * POP3Server listens on port 110 for connections and starts new POP3Thread for each incomming request
 */
class POP3Server extends Thread {
	static final int PORT = 110;
	HashMap users;
	ServerSocket s;

	/**
	 * Constructor - opens the socket
	 *
	 * @param users - Users' mailboxes
	 */
	public POP3Server(HashMap users) {
		this.users = users;
		try {
			// open the port
			s = new ServerSocket(PORT);
			start();
		} catch (IOException e) {}
	}

	/**
	 * The POP3 server lifecycle
	 */
	public void run() {
		try {
			// listen for incomming connections
  			while (true) {
				Socket socket = s.accept();
				try {
					new POP3Thread(socket, users);
				} catch (IOException e) {
					socket.close();
				}
			}
		} catch (IOException e) {
		} finally {
			try {
                s.close();
            } catch(IOException e) {}
		}
	}

}

/**
 * Simple Mail Server
 * Author, Georgi Mungov 2002
 *
 * The main program. Creates users' mailboxes and starts the SMTP and POP3 servers
 */
public class MailServer {
	public static void main(String args[]) {
		HashMap users = new HashMap();

		// add the users
		users.put("user1", new Mailbox("user1"));
		users.put("user2", new Mailbox("user2"));
		users.put("user3", new Mailbox("user3"));

		// start SMTP
		new SMTPServer(users);

		// start POP3
		new POP3Server(users);

	}
}
