/*
	UDP Chat Client
	Internet Programming with Java Course
	
	(c) 2002 Martin Valkanov
	
	UDP Chat is a client/server message exchange system.
	The user can send messages directly to a specific other user,
	or can use the server to broadcast a message to all users
	currently registered to the server.
	
	The main thread reads input from the console and implements the
	sending part.
	
	The other thread UDPClientListener receives messages from the server
	or other clients and displays them to the console.
	
	Usage: UDPChatClient [-serverhost=<IP_ADDRESS|HOSTNAME>] [-serverport=<server listening port>] [-localport=<client listening port>]
	Notes:
		If you do not specify local port, the OS will choose the first free one to be used.
		Default server address is 127.0.0.1.
		Default server port is 2002.
		
	Compiled with JDK 1.3.0
*/

import java.net.InetAddress;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.StringTokenizer;

public class UDPChatClient
{
	final static int DEFAULT_PORT = 2002;
	private InetAddress mServerAddr;
	private int mServerPort;
	private int mLocalPort;
	
	private	DatagramSocket mSock = null;
	private BufferedReader mIn = null;
	private PrintWriter mOut = null;
	
	private UDPClientListener mListener = null;
	
	public UDPChatClient( InetAddress serverAddr, int serverPort, int localPort )
	{
		mServerAddr = serverAddr;
		mServerPort = serverPort;
		mLocalPort = localPort;
	}
	
	// Create the datagram socket to be used for communication
	boolean initSocket()
	{
		try
		{
			// If the user did not specify listening port let the OS choose it.
			if ( mLocalPort == 0 )
				mSock = new DatagramSocket();
			else
			{
				try { mSock = new DatagramSocket( mLocalPort );	}
				catch ( SecurityException exSecur )
				{
					System.err.println( "Unable to create and bind socket." );
					System.err.println( "Sercurity exception: " + exSecur );
					exSecur.printStackTrace();					
				}
			}
			
			return true;
		}
		catch ( SocketException exSock )
		{
			System.err.println( "Unable to create socket." );
			System.err.println( "Socket exception: " + exSock );
			exSock.printStackTrace();
		}
		
		return false;
	}
	
	boolean initConsoleStreams()
	{
		mIn = new BufferedReader( new InputStreamReader( System.in ) );
		mOut = new PrintWriter( new OutputStreamWriter( System.out ) );
		return true;
	}
	
	// Send a single message to the server.
	void sendServerMsg( String sMsg ) throws IOException
	{
		byte buf[] = sMsg.getBytes();
		DatagramPacket dp = new DatagramPacket( buf, buf.length, mServerAddr, mServerPort );
		
		mSock.send( dp );
	}
	
	// Send a single message to a specific user.
	void sendClientMsg( String sMsg, InetAddress recipientAddr, int port ) throws IOException
	{
		byte buf[] = sMsg.getBytes();
		DatagramPacket dp = new DatagramPacket( buf, buf.length, recipientAddr, port );
		
		mSock.send( dp );
	}
	
	void joinGroup() throws IOException
	{
		sendServerMsg( "REGISTER" );
	}	
	
	void leaveGroup() throws IOException
	{
		sendServerMsg( "UNREGISTER" );
	}
	
	void registerName( String sName ) throws IOException
	{
		sendServerMsg( "NAME " + sName );
	}
	
	// After initialization print a welcome message to the user to indicate 
	// everything is OK.
	void printWelcomeMsg()
	{
		try
		{
			mOut.println( "Welcome to UDP chat" );
			String sWelcomeMsg = "UDP Chat Client started @ " + InetAddress.getLocalHost().getHostAddress() + ":" + mSock.getLocalPort();
			char sLine[] = new char[sWelcomeMsg.length()];
			for ( int i = 0; i < sLine.length; i++ ) sLine[i] = '-';
			mOut.println( sWelcomeMsg );
			mOut.println( sLine );
			mOut.flush();
		}
		catch ( UnknownHostException exUnknownHost )
		{
		}
	}
	
	// Detect syntax errors and display appropriate help to the user.
	class IllegalSyntaxException extends Throwable
	{
	}
	
	void illegalSyntax()
	{
		mOut.println( "Illegal command/command syntax. Valid commands are:" );
		mOut.println( "\tSEND <IP address>[:PORT] <message>" );
		mOut.println( "\tSEND * <message>" );
		mOut.println( "\tSEND <nickname> <message>" );
		mOut.println( "\tNAME <nickname>" );
		mOut.println( "\tQUIT" );
		mOut.flush();
	}
	
	// Parse the command typed by the user
	// If QUIT, instruct the client to finish by returning true.
	boolean parseAndExecuteCommand( String sCmd ) throws IllegalSyntaxException
	{
		StringTokenizer strTok = new StringTokenizer( sCmd, " ", false );
		if ( !strTok.hasMoreTokens() )
			throw new IllegalSyntaxException();

		String sCode = strTok.nextToken().toUpperCase();
		if ( sCode.equals( "QUIT" ) )
		{
			try { leaveGroup(); }
			catch ( IOException exIO )
			{
				System.err.println( "IOException: " + exIO );
				exIO.printStackTrace();
			}
			return true;
		}

		if ( sCode.equals( "NAME" ) )
		{
			String sName = strTok.nextToken( "\n" ).trim();
			if ( sName.length() == 0 )
				throw new IllegalSyntaxException();
			
			try { registerName( sName ); }
			catch ( IOException exIO )
			{
				System.err.println( "IOException: " + exIO );
				exIO.printStackTrace();
			}
		}
		else
		if ( sCode.equals( "SEND" ) )
		{
			if ( !strTok.hasMoreTokens() )
				throw new IllegalSyntaxException();
			
			try
			{
				String sRecipient = strTok.nextToken();
				char cFirstLetter = ' ';
				
				if ( sRecipient.length() != 0 )
					cFirstLetter = sRecipient.substring( 0, 1 ).toUpperCase().charAt(0);
				
				if ( sRecipient.equals( "*" ) ||
					( ( cFirstLetter >= 'A' ) && ( cFirstLetter <= 'Z' ) ) )
					sendServerMsg( sCmd );
				else
				{
					StringTokenizer strRecipientTok = new StringTokenizer( sRecipient, ":", false );
					String sRecipientHost = strRecipientTok.nextToken();
					int nPort;
					if ( strRecipientTok.hasMoreTokens() )
					{
						String sPort = strRecipientTok.nextToken();
						try	{ nPort = Integer.parseInt(sPort); }
						catch ( NumberFormatException exNumber )
						{
							throw new IllegalSyntaxException();
						}
					}
					else
						nPort = DEFAULT_PORT;
					
					try
					{
						InetAddress recipientAddr = InetAddress.getByName( sRecipientHost );
						String sMsg = strTok.nextToken( "\n" ).trim();
						sendClientMsg( sMsg, recipientAddr, nPort );
					}
					catch ( UnknownHostException exUnknownHost )
					{
						System.err.println( "Host " + sRecipientHost + " unknown." );
					}
				}
			}
			catch ( IOException exIO )
			{
				System.err.println( "IOException: " + exIO );
				exIO.printStackTrace();
			}
		}
		else
			throw new IllegalSyntaxException();
		
		return false;
	}
	
	// Starts the listening thread
	void startListener()
	{
		mListener = new UDPClientListener( mSock, mOut );
		mListener.start();
	}
	
	public void run()
	{
		// Establish the datagram socket.
		// If not successful, leave the program.
		if ( !initSocket() ) return;
		if ( !initConsoleStreams() ) { mSock.close(); return; }

		// Register in the server
		try { joinGroup(); }
		catch ( IOException exIO )
		{
			System.err.println( "Unable to join." );
			System.err.println( "IOException: " + exIO );
			exIO.printStackTrace();
			return;
		}

		printWelcomeMsg(); // Tell the user initialization is OK.
		startListener(); // Begin receiving messages.
		
		String sCmd;
		
		while (true)
		{
			// Read a command from the console.
			try { sCmd = mIn.readLine(); }
			catch ( IOException exIO )
			{
				System.err.println( "IOException: " + exIO );
				exIO.printStackTrace();
				break;
			}
			
			if ( sCmd == null )
			{
				try { leaveGroup(); }
				catch ( IOException exIO )
				{
					System.err.println( "IOException: " + exIO );
					exIO.printStackTrace();
				}				
				break;
			}
		
			// Recognize and carry out the command.
			try
			{
				if ( parseAndExecuteCommand( sCmd ) ) break;
			}
			catch ( IllegalSyntaxException exIllegalSyntax )
			{
				illegalSyntax();
			}
		}
		
		// Shutdown the receiving thread and the datagram socket.
		mListener.interrupt();
		mSock.close();
	}
	
	public static void main( String args[] )
	{
		String sServerHost = "127.0.0.1";
		int nServerPort = DEFAULT_PORT;
		int nLocalPort = 0;
		
		// Handle command line arguments
		// Let the user choose where the server is and where the client should listen.
		for ( int i = 0; i < args.length; i++ )
		{
			if ( args[i].toLowerCase().equals( "-help" ) ||	args[i].equals( "/?" ) )
			{
				System.out.println( "Usage: UDPChatClient [-serverhost=<IP_ADDRESS|HOSTNAME>] [-serverport=<server listening port>] [-localport=<client listening port>]" );
				System.out.println( "Notes:" );
				System.out.println( "\tIf you do not specify local port, the OS will choose the first free one to be used." );
				System.out.println( "\tDefault server address is 127.0.0.1." );
				System.out.println( "\tDefault server port is 2002." );
				return;
			}
			else
			if ( args[i].toLowerCase().startsWith( "-serverhost=" ) )
				sServerHost = args[i].substring( "-serverhost=".length() );
			else
			if ( args[i].toLowerCase().startsWith( "-serverport=" ) )
			{
				try
				{
					nServerPort = Integer.parseInt( args[i].substring( "-serverport=".length() ) );
				}
				catch ( NumberFormatException exNumberFormat ) {}
			}
			else
			if ( args[i].toLowerCase().startsWith( "-localport=" ) )
			{
				try
				{
					nLocalPort = Integer.parseInt( args[i].substring( "-localport=".length() ) );
				}
				catch ( NumberFormatException exNumberFormat ) {}
			}
		}

		// Startup the chat client passing it server address, server port and
		// local listening port.
		try
		{	
			InetAddress serverAddr = InetAddress.getByName( sServerHost );
			UDPChatClient client = new UDPChatClient( serverAddr, nServerPort, nLocalPort );
			client.run();
		}
		catch ( UnknownHostException exUnknownHost )
		{
			System.err.println( "Can not access server. Unknown host: " + sServerHost );
			System.err.println( "UnknownHostException: " + exUnknownHost );
			exUnknownHost.printStackTrace();
		}
	}
};