/*
	UDP Chat Server
	Internet Programming with Java Course
	
	(c) 2002 Martin Valkanov
		
	The ServerDispatcher maintains a list of clients and broadcasts
	queued messages to all clients through their ClientServer threads.
	
	It handles client registering/unregistering.
*/

import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.net.UnknownHostException;

public class ServerDispatcher extends Thread
{
	private Vector mClientList = new Vector();
	private Hashtable mClientNameTable = new Hashtable();
	private Vector mMessageQueue = new Vector();
	private boolean mNotTerminated = true;

	// Register a client via pair of InetAddress and port.
	// Start its ClientSender thread.
	public void addClient( String clientId )
	{
		ClientInfo clientInfo;
		
		try
		{
			try
			{
				clientInfo = new ClientInfo( clientId );
			}
			catch ( NumberFormatException exNumberFormat )
			{
				return;
			}
		} catch ( UnknownHostException exUnknownHost )
		{
			return;
		}
		
		ClientSender sender = new ClientSender( clientInfo );
		clientInfo.mSender = sender;
		
		mClientList.add( clientInfo );
		clientInfo.mSender.start();
	}
	
	// Unregister a client. Terminate its sender thread.
	public void removeClient( String clientId )
	{
		int idx = -1;
		for ( int i = 0; i < mClientList.size(); i++ )
			if ( mClientList.elementAt(i).equals( clientId ) )
			{
				idx = i;
				break;
			}
			
		if ( idx != -1 )
		{
			ClientInfo clientInfo = (ClientInfo) mClientList.elementAt(idx);
			clientInfo.mSender.terminate();
			clientInfo.mSender.interrupt();
			mClientList.removeElementAt(idx);
		}
	}
	
	public void registerName( String sName, String clientId )
	{
		mClientNameTable.put( sName, clientId );
	}
	
	// Display server load status to the console.
	public void printClientList()
	{
		System.out.println( "" + mClientList.size() + " client(s) currently registered" + ( ( mClientList.size() > 0 ) ? ":" : "" ) );
		for ( int i = 0; i < mClientList.size(); i++ )
			System.out.println( ( (ClientInfo) mClientList.elementAt(i) ).mId );
	}
	
	// Pushback a message to the broadcasting queue.
	public void dispatchMessage( String clientId, String sMsg )
	{
		System.out.println( clientId + ": " + sMsg );
		if ( sMsg.toUpperCase().equals( "REGISTER" ) )
			addClient( clientId );
		else
		if ( sMsg.toUpperCase().equals( "UNREGISTER" ) )
			removeClient( clientId );
		else
		if ( sMsg.toUpperCase().startsWith( "NAME " ) )
			registerName( sMsg.substring( "NAME ".length() ), clientId );
		else
		synchronized ( mMessageQueue )
		{
			if ( sMsg.toUpperCase().startsWith( "SEND * " ) )
				sMsg = sMsg.substring( "SEND * ".length() );
			else
			if ( sMsg.toUpperCase().startsWith( "SEND " ) )
			{
				sMsg = sMsg.substring( "SEND ".length() );
				StringTokenizer strTok = new StringTokenizer( sMsg, " ", false );
				String sRecipientName = "";
				if ( strTok.hasMoreTokens() )
				{
					sRecipientName = strTok.nextToken();
					sMsg = strTok.nextToken( "\n" );

					sendMessageToNamedUser( sRecipientName, "Received message from " + clientId + ":" + sMsg );
				}
				
				return;
			}
			
			mMessageQueue.add( "Received message from " + clientId + ": " + sMsg );
			mMessageQueue.notify();
		}
	}
	
	// Sends a message to a single user identified by name.
	void sendMessageToNamedUser( String sRecipientName, String sMsg )
	{
		String clientId = (String) mClientNameTable.get( sRecipientName );
		if ( clientId != null )
		{			
			int idx = -1;
			for ( int i = 0; i < mClientList.size(); i++ )
				if ( mClientList.elementAt(i).equals( clientId ) )
				{
					idx = i;
					break;
				}
				
			if ( idx != -1 )
			{
				ClientInfo clientInfo = (ClientInfo) mClientList.elementAt(idx);
				clientInfo.mSender.queueMessage( sMsg );
			}
		}
	}
	
	// Retrieve a message to be broadcasted.
	// If the queue is empty, wait.
	String getNextMessage() throws InterruptedException
	{
		while ( mMessageQueue.isEmpty() )
			synchronized ( mMessageQueue ) { mMessageQueue.wait(); }
			
		String sMsg;
		
		synchronized ( mMessageQueue )
		{
			sMsg = (String) mMessageQueue.elementAt(0);
			mMessageQueue.remove(0);
		}
		
		return sMsg;		
	}
	
	// Deliver the message to all client sender threads.
	void sendMessageToAllClients( String sMsg )
	{
		for ( int i = 0; i < mClientList.size(); i++ )
		{
			ClientInfo client = (ClientInfo) mClientList.elementAt(i);
			client.mSender.queueMessage( sMsg );
		}
	}
	
	// Cleanup the client list and terminate the sender threads.
	public void terminate()
	{
		mNotTerminated = false;
		for ( int i = 0; i < mClientList.size(); i++ )
		{
			ClientInfo client = (ClientInfo) mClientList.elementAt(i);
			client.mSender.interrupt();
		}
	}
	
	public void run()
	{
		try
		{
			// Message pump
			while ( mNotTerminated )
			{
				String sMsg = getNextMessage();
				sendMessageToAllClients( sMsg );
			}
		}
		catch ( InterruptedException exIntr )
		{
		}
	}
};