/*
	UDP Chat Server
	Internet Programming with Java Course
	
	(c) 2002 Martin Valkanov
	
	UDP Chat is a client/server message exchange system.
	The user can send messages directly to a specific other user,
	or can use the server to broadcast a message to all users
	currently registered to the server.
	
	The main server thread listens for messages and commands from the
	clients and redirects them to the ServerDispatcher thread.
	
	The ServerDispatcher maintains a list of clients and broadcasts
	queued messages to all clients through their ClientServer threads.
	
	Usage: UDPChatServer [-port=<listening port number>]
	
	Valid server commands are: clientlist, quit.
	
	Complied with JDK 1.3.0
*/

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.net.InetAddress;
import java.net.UnknownHostException;

public class UDPChatServer extends Thread
{
	private final static int DEFAULT_PORT = 2002;
	private final static int BUF_SIZE = 512;
	private int mPort; 				// UDPClient shall contact us here.
	private DatagramSocket mSocket; // Message/command listener
	private ServerDispatcher mDispatcher;	// Message broadcaster
	private boolean mNotTerminated = true;	// Used to instruct the server to stop
											// when the server user types QUIT.
	
	UDPChatServer( int port ) throws SocketException
	{
		mPort = port;
		mSocket = new DatagramSocket( mPort );		
	}
	
	public void terminate()
	{
		mNotTerminated = false;
		mSocket.close();
	}
	
	public void run()
	{
		DatagramPacket dp;
		byte buf[] = new byte[BUF_SIZE];
		
		// Tell the user the server initialization is OK.
		try
		{
			String sWelcomeMsg = "UDP Chat Server started @ " + InetAddress.getLocalHost().getHostAddress() + ":" + mSocket.getLocalPort();
			char sLine[] = new char[sWelcomeMsg.length()];
			for ( int i = 0; i < sLine.length; i++ ) sLine[i] = '-';
			System.out.println( sWelcomeMsg );
			System.out.println( sLine );
		}
		catch ( UnknownHostException exUnknownHost )
		{
		}
		
		// Start the broadcaster		
		mDispatcher = new ServerDispatcher();
		mDispatcher.start();

		// Message/command receiving loop		
		while ( mNotTerminated )
		{
			dp = new DatagramPacket( buf, buf.length );
			try
			{
				mSocket.receive(dp);
				mDispatcher.dispatchMessage( dp.getAddress().getHostAddress() + ":" + dp.getPort(), new String( dp.getData(), dp.getOffset(), dp.getLength() ) );
			}
			catch ( IOException exIO )
			{
				if ( mNotTerminated )
				{
					System.err.println( "IOException: " + exIO );
					exIO.printStackTrace();
				}
				else
				{
					System.err.println( "Terminated gracefully." );
				}
			}			
		}
		
		// Shutdown the broadcaster and all client senders
		mDispatcher.terminate();
		mDispatcher.interrupt();
	}
	
	public static void main( String args[] )
	{
		UDPChatServer server;
		int nPort = DEFAULT_PORT;
		
		// Use a default listening port unless the user specify one.
		for ( int i = 0; i < args.length; i++ )
		{
			if ( args[i].toLowerCase().equals( "-help" ) ||
				args[i].equals( "/?" ) )
			{
				System.out.println( "Usage: UDPChatServer [-port=<listening port number>]" );
				return;
			}
			else
			if ( args[i].toLowerCase().startsWith( "-port=" ) )
			try
			{
				nPort = Integer.parseInt( args[i].substring( "-port=".length() ) );
			}
			catch ( NumberFormatException exNumberFormat )
			{
			}
		}

		// Startup the server		
		try
		{
			server = new UDPChatServer( nPort );			
		}
		catch ( SocketException exSocket )
		{
			System.err.println( "Server socket creation failed: " + exSocket );
			exSocket.printStackTrace();
			return;
		}

		server.start();
		
		// Let the user have control of the server.
		BufferedReader in = new BufferedReader( new InputStreamReader( System.in ) );
		String s;
		try
		{
			while ( ( s = in.readLine() ) != null )
			{
				if ( s.toUpperCase().equals( "HELP" ) )
					System.out.println( "Valid commands are: QUIT, CLIENTLIST and HELP." );
				else
				if ( s.toUpperCase().equals( "QUIT" ) )
					break;
				else
				if ( s.toUpperCase().equals( "CLIENTLIST" ) )
					server.mDispatcher.printClientList();
			}
			server.terminate();
		}
		catch ( IOException exIO )
		{
			System.err.println( "IOException: " + exIO );
			exIO.printStackTrace();
		}
	}
};