package fileclient;

/*
 * Created by IntelliJ IDEA.
 * User: peter
 * Date: Apr 17, 2002
 * Time: 3:21:29 PM
 */

import utils.*;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * File client class. Follows the file transfer protocol, as specified by Svetlin Nakov
 */
public class FileClient extends Thread {
	static final int DEFAULT_PORT = 2002;
	static final int BLOCK_SIZE = 65536;

	private Socket socket;
	private InputStream inStream;
	private OutputStream outStream;
	private CustomSocketReader inReader;
	private CustomSocketWriter outWriter;
	private BufferedReader stdIn;
	private CommandParser parser;
	private String home = "";

	/**
	 * @param host The host to connect to
	 * @param port The port to use
	 * @throws IOException If the host is found, but a connection cannot be established
	 * @throws UnknownHostException If the host cacont be connected
	 */
	FileClient(String host, int port) throws IOException, UnknownHostException {
		parser = new CommandParser();
		parser.init("clientcommands.properties");

		InetAddress addr;
		addr = InetAddress.getByName(host);
		try {
			socket = new Socket(addr, port);
			System.out.println("Connected to server on " + socket);

			inStream = socket.getInputStream();
			outStream = socket.getOutputStream();
			inReader = new CustomSocketReader(inStream);
			outWriter = new CustomSocketWriter(outStream);
			stdIn = new BufferedReader(new InputStreamReader(System.in));

			Properties props = new Properties();
			try {
				props.load(new FileInputStream("client.properties"));
				home = props.getProperty("home_directory");
				if (home == null) {
					File f = new File("");
					home = f.getAbsolutePath();
					System.out.println("User home not specified in the properties file: home_directory=<path>");
					System.out.println("Using default home directory " + home);
				}
			} catch (IOException e) {
				File f = new File("");
				home = f.getAbsolutePath();
				System.out.println("Properties file client.properties missing. Using default home directory " + home);
			}
			File homeFile = new File(home);
			if (!homeFile.exists() || !homeFile.isDirectory()) {
				File f = new File("");
				home = f.getAbsolutePath();
				System.out.println("Home directory does not exist. Using default home directory " + home);
			}
			if (!home.endsWith(File.separator) && !home.endsWith("/") && home.length() > 0)
				home = home + File.separator;

			start();

		} catch (IOException e) {
			System.out.println("Error: " + e.getMessage());
			if (socket != null)
				socket.close();
		}
	}

	/**
	 * Called by start in the constructor
	 * Reads commands from the command line and executes them
	 */
	public void run() {
		try {
			while (true) {
				displayPrompt();
				String command;
				try {
					command = readCommand();
				} catch (IOException e) {
					e.printStackTrace();
					break;
				}
				try {
					int result = executeCommand(command);
					if (result == -1) {
						break;
					}
				} catch (IOException e) {
					System.out.println("Could not execute command: " + e.getMessage());
					System.out.println("Client will exit");
					break;
				} catch (InvocationTargetException e) {
					System.out.println("Could not execute command: " + e.getMessage());
					System.out.println("Client will exit");
					break;
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				}
			}
		} finally {
			try {
				if (socket != null) {
					System.out.println("Closing " + socket);
					socket.close();
				}
			} catch (IOException e) {
				System.out.println("Could not close socket");
				e.printStackTrace();
			}
		}
	}

	/**
	 * Displays the ">" prompt
	 */
	protected void displayPrompt() {
		System.out.print(">");
	}

	/**
	 * Reads a command from the command line
	 * @return The command, trimmed
	 * @throws IOException
	 */
	protected String readCommand() throws IOException {
		if (stdIn == null) {
			stdIn =
				new BufferedReader(
					new InputStreamReader(System.in));
		}
		return stdIn.readLine().trim();
	}

	/**
	 * Uses reflection, based on the data, collected from the properties file to execute
	 * the necessary method, which handles the given command.
	 * @param command
	 * @return -1, if the client should quit, or 0 in all other cases
	 * @throws IllegalArgumentException Indicates an error in the parsing of the command properties.
	 * @throws InvocationTargetException In case an exception is throws by the executed method. This will most probably be an IOException
	 * @throws IOException
	 */
	protected int executeCommand(String command)
		throws IllegalArgumentException, InvocationTargetException, IOException {
		if (!parser.isLegalCommand(command)) {
			singlelineRequest(command);
			return 0;
		}
		CommandTokenizer tokens = new CommandTokenizer(command);
		String cmd = tokens.nextToken();
		int num = parser.getNumberOfArguments(cmd);
		Object[] arguments = new String[num + 1];
		arguments[0] = command.trim();
		for (int i = 1; i <= num; i++)
			arguments[i] = tokens.nextToken();
		Class[] argClasses;
		if (num > 0) {
			argClasses = new Class[num + 1];
			for (int i = 0; i < argClasses.length; i++)
				argClasses[i] = String.class;
		} else {
			argClasses = new Class[1];
			argClasses[0] = String.class;
		}
		String methodName = parser.getMethodName(cmd);
		Method method;
		try {
			method = this.getClass().getMethod(methodName, argClasses);
			Integer result = (Integer) method.invoke(this, arguments);
			return result.intValue();
		} catch (NoSuchMethodException e) {
			singlelineRequest(command);
			return 0;
		} catch (IllegalAccessException e) {
			singlelineRequest(command);
			return 0;
		} catch (SecurityException e) {
			singlelineRequest(command);
			return 0;
		}
	}

	/**
	 * Main method
	 * @param args
	 */
	public static void main(String[] args) {
		String host = "localhost";
		int port = DEFAULT_PORT;

		if (args.length > 0) {
			host = args[0];
			if (args.length > 1) {
				try {
					port = Integer.parseInt(args[1]);
				} catch (NumberFormatException e) {
					System.out.println("Illegal port number");
				}
			}
		}

		try {
			new FileClient(host, port);
		} catch (UnknownHostException e) {
			System.out.println("Host unknown");
		} catch (IOException e) {
			System.out.println(e.getMessage());
		}
	}

	/**
	 * Sends the given command to the server and expetcs a one-line answer from it
	 * @param command
	 * @return The response of the server
	 * @throws IOException
	 */
	private String singlelineRequest(String command) throws IOException {
		outWriter.println(command.trim());
		String result = inReader.readln();
		System.out.println(result);
		return result;
	}

	/**
	 * Sends the given command to the server and expetcs a multi-line answer ending with a dot from it.
	 * If the answer starts with a line, starting with "-", the answer is considered to indicate an error and
	 * is therefore considered to be single-line.
	 * @param command
	 * @return The first line from the response, which indicates error or success of the request.
	 * @throws IOException
	 */
	private String multilineRequest(String command) throws IOException {
		outWriter.println(command);
		String line;
		String result = inReader.readln();
		System.out.println(result);
		if (result.trim().startsWith(".") || result.trim().startsWith("-"))
			return result;
		do {
			line = inReader.readln().trim();
			System.out.println(line);
		} while (!line.equals("."));
		return result;
	}

	/**
	 * Called on commands for exiting
	 * @param cmd
	 * @return -1, so as the thread to stop and close the socket
	 * @throws IOException
	 */
	public int quit(String cmd) throws IOException {
		String result = singlelineRequest(cmd);
		if (result.startsWith("+"))
			return -1;
		else
			return 0;
	}

	public int list(String cmd) throws IOException {
		multilineRequest(cmd);
		return 0;
	}

	public int chdir(String cmd, String newdir) throws IOException {
		singlelineRequest(cmd);
		return 0;
	}

	public int pwd(String cmd) throws IOException {
		singlelineRequest(cmd);
		return 0;
	}

	public int help(String cmd) throws IOException {
		System.out.println("Valid local commands are: ");
		System.out.println("_DIR - Lists the contents of the local home directory");
		System.out.println();
		multilineRequest(cmd);
		return 0;
	}

	/**
	 * User friendly form to send an existing file
	 * @param cmd
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public int upload(String cmd, String filename) throws IOException {
		File file = new File(home + filename);
		String response;
		if (!file.exists() || !file.isFile()) {
			System.out.println("Cannot find file " + filename);
			return 0;
		}
		FileInputStream finStr;
		try {
			finStr = new FileInputStream(file);
		} catch (FileNotFoundException e) {
			System.out.println("Cannot find file " + filename);
			return 0;
		}
		try {
			outWriter.println(cmd);
			response = inReader.readln();
			if (response.startsWith("+")) {
				System.out.println("Sending file " + filename);
				long len = file.length();
				outWriter.println((new Long(len)).toString());
				byte[] arr = new byte[BLOCK_SIZE];
				for (long i = 0; i < len;) {
					int read = finStr.read(arr);
					i += read;
					outWriter.write(arr, 0, read);
				}
				System.out.println(inReader.readln());
			} else {
				System.out.println(response);
			}
			return 0;
		} finally {
			finStr.close();
		}
	}

	/**
	 * Sends a file, as defined by the protocol. (length, contents entered by the user)
	 * @param cmd
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public int send(String cmd, String filename) throws IOException {
		outWriter.println(cmd);
		String response = inReader.readln();
		if (response.startsWith("+")) {
			System.out.println(response);
			long len = new Long(stdIn.readLine().trim()).longValue();
			outWriter.println((new Long(len)).toString());
			char[] arr = new char[BLOCK_SIZE];
			for (long i = 0; i < len;) {
				int read = stdIn.read((char[]) arr, 0, (int) Math.min(len - i, BLOCK_SIZE));
				i += read;
				if (i > len) {
					read = read - (int) (i - len);
				}
				byte[] hlp = new byte[read];
				for (int j = 0; j < read; j++) {
					hlp[j] = (byte) arr[j];
				}
				outWriter.write(hlp, 0, read);
			}
			while (stdIn.ready()) {
				stdIn.skip(1);
			}
			System.out.println(inReader.readln());
		} else {
			System.out.println(response);
		}
		return 0;
	}

	/**
	 * Gets a file from the server
	 * @param cmd
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public int get(String cmd, String filename) throws IOException {
		File file = new File(home + filename);
		if (file.exists()) {
			System.out.println("Cannot store file in current directory: File exists");
			return 0;
		}
		FileOutputStream foutStr;
		try {
			foutStr = new FileOutputStream(home + filename);
		} catch (FileNotFoundException e) {
			System.out.println("Cannot open locally file " + filename + " for writing");
			return 0;
		}
		try {
			outWriter.println(cmd);
			String response;
			response = inReader.readln();
			if (response.startsWith("+")) {
				long len = (new Long(inReader.readln())).longValue();
				byte[] arr = new byte[BLOCK_SIZE];
				for (long i = 0; i < len;) {
					int read = inReader.read(arr, 0, (int) Math.min(len - i, BLOCK_SIZE));
					i += read;
					foutStr.write(arr, 0, read);
				}
				System.out.println("File " + filename + " stored");
			} else {
				foutStr.close();
				file = new File(home + filename);
				try {
					if (file.exists())
						file.delete();
				} catch (Exception e) {
					e.printStackTrace();
				}
				System.out.println(response);
			}
		} finally {
			foutStr.close();
		}
		return 0;
	}

	/**
	 * Lists the local home directory
	 * @param cmd
	 * @return
	 * @throws IOException
	 */
	public int localList(String cmd) throws IOException {
		final int OFFSET = 6;
		File dir = new File(home);
		if (!dir.exists() || !dir.isDirectory()) {
			System.out.println("Home directory is in illegal state. Client will close");
			quit(cmd);
			return -1;
		}
		System.out.println("Listing local home directory " + home);
		System.out.println();
		File[] listing = dir.listFiles();
		int maxlen = 4;
		int numfiles = 0;
		int numdirs = 0;
		int sizefiles = 0;
		ArrayList dirs = new ArrayList();
		ArrayList files = new ArrayList();
		for (int i = 0; i < listing.length; i++) {
			maxlen = Math.max(maxlen, listing[i].getName().length());
			if (listing[i].isDirectory()) {
				numdirs++;
				dirs.add(listing[i]);
			} else if (listing[i].isFile()) {
				numfiles++;
				sizefiles += listing[i].length();
				files.add(listing[i]);
			}
		}
		String line;
		if (home.length() > 3) {
			line = "..";
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + "<DIR>";
			System.out.println(line);
		}
		for (Iterator it = dirs.iterator(); it.hasNext();) {
			File curDir = (File) it.next();
			line = curDir.getName();
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + "<DIR>";
			System.out.println(line);
		}
		for (Iterator it = files.iterator(); it.hasNext();) {
			File curFile = (File) it.next();
			line = curFile.getName();
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + curFile.length();
			System.out.println(line);
		}
		System.out.println();
		System.out.println("Total " + numfiles + " files (" + sizefiles + " bytes), " + numdirs + " subdirectories.");
		System.out.println();

		return 0;
	}

}
