package fileserver;

/*
 * Created by IntelliJ IDEA.
 * User: peter
 * Date: Apr 17, 2002
 * Time: 2:52:20 PM
 */

import utils.*;

import java.io.*;
import java.net.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * An object from this class is created by the main server class to serve a single client
 */
class SingleFileServer extends Thread {
	static final int BLOCK_SIZE = 65536;

	private static CommandParser parser;
	private static String home = "";

	private Socket socket;
	private InputStream inStream;
	private OutputStream outStream;
	private CustomSocketReader inReader;
	private CustomSocketWriter outWriter;
	private String currentDir = "/";

	public SingleFileServer(Socket socket, CommandParser parser, String home) throws IOException {

		this.socket = socket;
		this.parser = parser;
		this.home = home;

		inStream = socket.getInputStream();
		outStream = socket.getOutputStream();
		inReader = new CustomSocketReader(inStream);
		outWriter = new CustomSocketWriter(outStream);

		currentDir = "/";

		start(); // Calls run()
	}

	public void run() {
		try {
			while (true) {
				String command;
				try {
					command = readCommand();
				} catch (IOException e) {
					System.out.println(e.getMessage());
					break;
				} catch (Exception e) {
					e.printStackTrace();
					break;
				}
				try {
					try {
						int result = executeCommand(command);
						if (result == -1)
							break;
					} catch (NoSuchMethodException e) {
						outWriter.println("-Command is legal but method is not implemented yet");
					} catch (IllegalAccessException e) {
						outWriter.println("-Command is legal but no access rights to execute method");
					} catch (IOException e) {
						System.out.println("Error: " + e.getMessage());
						break;
					} catch (Exception e) {
						outWriter.println("-" + e.getMessage());
						e.printStackTrace();
					}
				} catch (IOException e) {
					e.printStackTrace();
					break;
				}
			}
		} finally {
			try {
				if (socket != null) {
					System.out.println("Disconnecting client on " + socket);
					socket.close();
				}
			} catch (IOException e) {
				System.out.println("Could not close socket");
				e.printStackTrace();
			}
		}
	}

	/**
	 * Uses reflection to execute the necessary method for a certain command
	 * All exceptions are caught in the run method and handled appropriately
	 * @param command
	 * @return
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws IOException
	 */
	private int executeCommand(String command)
		throws NoSuchMethodException, SecurityException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, IOException {
		if (!parser.isLegalCommand(command)) {
			outWriter.println("-Invalid command");
			return 0;
		}
		CommandTokenizer tokens = new CommandTokenizer(command);
		String cmd = tokens.nextToken();
		int num = parser.getNumberOfArguments(cmd);
		Object[] arguments = new String[num + 1];
		arguments[0] = cmd.toUpperCase().trim();
		for (int i = 1; i <= num; i++)
			arguments[i] = tokens.nextToken();
		Class[] argClasses;
		if (num > 0) {
			argClasses = new Class[num + 1];
			for (int i = 0; i < argClasses.length; i++)
				argClasses[i] = String.class;
		} else {
			argClasses = new Class[1];
			argClasses[0] = String.class;
		}
		String methodName = parser.getMethodName(cmd);
		Method method;
		method = this.getClass().getMethod(methodName, argClasses);
		Integer result = (Integer) method.invoke(this, arguments);
		return result.intValue();
	}

	/**
	 * Reads a command from the client
	 * @return
	 * @throws IOException
	 */
	protected String readCommand() throws IOException {
		return inReader.readln();
	}

	/**
	 * Disconnects the client
	 * @param cmd
	 * @return Always -1 so as the thread to stop and close the socket
	 * @throws IOException
	 */
	public int quit(String cmd) throws IOException {
		outWriter.println("+Bye");
		return -1;
	}

	/**
	 * This method handles a GET command from the client
	 * @param cmd
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public int get(String cmd, String filename) throws IOException {
		String filePath = home + currentDir;
		File file = new File(filePath + filename);
		if (!file.exists() || !file.isFile()) {
			outWriter.println("-Cannot find file " + filename);
			return 0;
		}
		FileInputStream finStr;
		try {
			finStr = new FileInputStream(file);
		} catch (FileNotFoundException e) {
			outWriter.println("-Cannot find file " + filename);
			return 0;
		}
		try {
			outWriter.println("+File size and file contents follow:");
			long len = file.length();
			outWriter.println((new Long(len)).toString());
			byte[] arr = new byte[BLOCK_SIZE];
			for (long i = 0; i < len;) {
				int read = finStr.read(arr);
				i += read;
				outWriter.write(arr, 0, read);
			}
			return 0;
		} finally {
			finStr.close();
		}
	}

	/**
	 * This method handles a send command from the client
	 * @param cmd
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public int send(String cmd, String filename) throws IOException {
		String filePath = home + currentDir;
		File file = new File(filePath + filename);
		FileOutputStream foutStr;
		if (file.exists()) {
			outWriter.println("-Cannot store file in current directory: File exists");
			return 0;
		}
		try {
			foutStr = new FileOutputStream(file);
		} catch (Exception e) {
			outWriter.println("-Cannot store file in current directory: " + e.getMessage());
			return 0;
		}
		try {
			outWriter.println("+Ready to accept file");
			long len = (new Long(inReader.readln().trim())).longValue();
			byte[] arr = new byte[BLOCK_SIZE];
			for (long i = 0; i < len;) {
				int read = inReader.read(arr, 0, (int) Math.min(len - i, BLOCK_SIZE));
				i += read;
				foutStr.write(arr, 0, read);
			}
			outWriter.println("+File " + filename + " stored");
		} finally {
			foutStr.close();
		}
		return 0;
	}

	/**
	 * Sends the help information to the client
	 * @param cmd
	 * @return
	 * @throws IOException
	 */
	public int help(String cmd) throws IOException {
		outWriter.println("+Valid commands are:");
		outWriter.println("PWD - shows current directory");
		outWriter.println("DIR - Lists current directory");
		outWriter.println("CD directory - Changes current directory");
		outWriter.println("GET filename - Retrieves given file from current directory");
		outWriter.println("SEND filename");
		outWriter.println("filesize");
		outWriter.println("<file contents> - Sends given file to the server");
		outWriter.println("HELP - Displays this help");
		outWriter.println("QUIT - Closes this session");
		outWriter.println(".");
		return 0;
	}

	private String getDisplayDirectory(File absDir) throws IOException {
		String absPath = Util.replace(absDir.getCanonicalPath(), File.separator, "/");
		String absHome = home;
		if (absHome.endsWith("/"))
			absHome = absHome.substring(0, absHome.length() - 1);
		absPath = absPath.substring(absHome.length());
		if (absPath.length() == 0)
			absPath = "/";
		else if (absPath.endsWith("/"))
			absPath = absPath.substring(0, absPath.length() - 1);
		return absPath;
	}

	private String getDisplayDirectory() throws IOException {
		File absDir = new File(home + currentDir);
		String absPath = Util.replace(absDir.getCanonicalPath(), File.separator, "/");
		String absHome = home;
		if (absHome.endsWith("/"))
			absHome = absHome.substring(0, absHome.length() - 1);
		absPath = absPath.substring(absHome.length());
		if (absPath.length() == 0)
			absPath = "/";
		else if (absPath.endsWith("/"))
			absPath = absPath.substring(0, absPath.length() - 1);
		return absPath;
	}

	/**
	 * Sends the current directory name to the client
	 * @param cmd
	 * @return
	 * @throws IOException
	 */
	public int pwd(String cmd) throws IOException {
		currentDir = Util.replace(currentDir, File.separator, "/");
		outWriter.println("+Current directory is " + getDisplayDirectory());
		return 0;
	}

	/**
	 * Handles the cd command
	 * @param cmd
	 * @param newdir It can be any relative, absolute (starting from / - home), complex path as long as it is a valid directory and is under the home directory
	 * @return
	 * @throws IOException
	 */
	public int chdir(String cmd, String newdir) throws IOException {
		newdir = newdir.trim();
		newdir = Util.replace(newdir, File.separator, "/");
		if (newdir.endsWith("/") && newdir.length() > 1)
			newdir = newdir.substring(0, newdir.length() - 1);
		if (newdir.startsWith("..") && currentDir.length() <= 1) {
			outWriter.println("-Invalid directory");
			return 0;
		}
		File test;
		if (newdir.startsWith("/"))
			test = new File(home + newdir);
		else
			test = new File(home + currentDir + newdir);
		if (!test.exists() || !test.isDirectory()) {
			outWriter.println("-Invalid directory");
			return 0;
		}
		String testPath = Util.replace(test.getCanonicalPath(), File.separator, "/").toLowerCase();
		String testHome = Util.replace(home, File.separator, "/").toLowerCase();
		if (!testHome.endsWith("/")) testHome = testHome + "/";
		if (!testPath.endsWith("/")) testPath = testPath + "/";
		if (!testPath.startsWith(testHome)) {
			outWriter.println("-Invalid directory");
			return 0;
		}
		String newCurDir = getDisplayDirectory(test);
		outWriter.println("+Directory changed to " + newCurDir);
		currentDir = newCurDir;
		if (!currentDir.endsWith("/"))
			currentDir = currentDir + "/";
		return 0;
	}

	/**
	 * Handles the dir command
	 * @param cmd
	 * @return
	 * @throws IOException
	 */
	public int list(String cmd) throws IOException {
		final int OFFSET = 6;
		File dir = new File(home + currentDir);
		if (!dir.exists() || !dir.isDirectory()) {
			outWriter.println("-Server directory is in illegal state. Server will close connection");
			return -1;
		}
		outWriter.println("+Listing directory " + getDisplayDirectory());
		outWriter.println("");
		File[] listing = dir.listFiles();
		int maxlen = 4;
		int numfiles = 0;
		int numdirs = 0;
		int sizefiles = 0;
		ArrayList dirs = new ArrayList();
		ArrayList files = new ArrayList();
		for (int i = 0; i < listing.length; i++) {
			maxlen = Math.max(maxlen, listing[i].getName().length());
			if (listing[i].isDirectory()) {
				numdirs++;
				dirs.add(listing[i]);
			} else if (listing[i].isFile()) {
				numfiles++;
				sizefiles += listing[i].length();
				files.add(listing[i]);
			}
		}
		String line;
		if (currentDir.length() > 1) {
			line = "..";
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + "<DIR>";
			outWriter.println(line);
		}
		for (Iterator it = dirs.iterator(); it.hasNext();) {
			File curDir = (File) it.next();
			line = curDir.getName();
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + "<DIR>";
			outWriter.println(line);
		}
		for (Iterator it = files.iterator(); it.hasNext();) {
			File curFile = (File) it.next();
			line = curFile.getName();
			while (line.length() < maxlen + OFFSET) line = line + " ";
			line = line + curFile.length();
			outWriter.println(line);
		}
		outWriter.println("");
		outWriter.println("Total " + numfiles + " files (" + sizefiles + " bytes), " + numdirs + " subdirectories.");
		outWriter.println(".");

		return 0;
	}
}

